// Copyright (C) 2023 WYATT GROUP
// Please see the AUTHORS file for details.
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program. If not, see <https://www.gnu.org/licenses/>.

import 'package:flutter/services.dart' show AssetBundle, rootBundle;
import 'package:wyatt_i18n/src/core/enums/format.dart';
import 'package:wyatt_i18n/src/core/exceptions/exceptions.dart';
import 'package:wyatt_i18n/src/core/utils/assets_utils.dart';
import 'package:wyatt_i18n/src/domain/data_sources/i18n_data_source.dart';

/// {@template assets_file_data_source_impl}
/// Implementation of [I18nDataSource] that loads i18n files from the assets.
///
/// The [basePath] is the folder where the i18n files are located.
/// The [baseName] is the name of the i18n files without the extension.
/// The [format] is the format of the i18n files.
///
/// For example, if the i18n files are located in the `assets/l10n/` and are
/// named `i18n.en.arb`, `i18n.fr.arb`, etc., then the [basePath]
/// is `l10n` and the [baseName] is `i18n` and the [format] is [Format.arb].
/// {@endtemplate}
class AssetsFileDataSourceImpl extends I18nDataSource {
  /// {@macro assets_file_data_source_impl}
  AssetsFileDataSourceImpl({
    this.basePath = '',
    this.baseName = 'i18n',
    super.format = Format.arb,
    super.defaultLocale = 'en',
  }) : super();

  /// The folder where the i18n files are located.
  ///
  /// Note: The path is relative to the `assets` folder. So, `assets/l10n/`
  /// is `l10n`.
  final String basePath;

  /// The name of the i18n files without the extension.
  final String baseName;

  /// The asset bundle used to load the i18n files.
  final AssetBundle assetBundle = rootBundle;

  /// Tries to load the i18n file from the given [locale].
  Future<String> _tryLoad(String? locale) async {
    String? content;
    final ext = format.name;

    final path = AssetsUtils.cleanPath(
      locale == null
          ? '$basePath/$baseName.$defaultLocale.$ext'
          : '$basePath/$baseName.$locale.$ext',
    );

    try {
      if (await AssetsUtils.assetExists(path)) {
        content = await assetBundle.loadString(path);
      }
    } catch (_) {
      content = null;
    }

    /// If the i18n file is not found, then we try to load the
    /// default i18n file.
    if (content == null && locale != null) {
      try {
        final fallbackPath = AssetsUtils.cleanPath(
          '$basePath/$baseName.$defaultLocale.$ext',
        );
        content = await assetBundle.loadString(fallbackPath);
      } catch (_) {
        throw SourceNotFoundException(path, format: format);
      }
    }

    /// If the default i18n file is not found, then we throw an exception.
    /// This case should happen only if the locale is null.
    if (content == null) {
      throw SourceNotFoundException(path, format: format);
    }

    return content;
  }

  /// Tries to load the i18n file from a given [uri].
  /// This method is used when the [uri] is not null.
  Future<String> _tryLoadUri(Uri uri) async {
    String? content;
    try {
      content = await assetBundle.loadString(uri.toString());
    } catch (_) {
      throw SourceNotFoundException(uri.toString(), format: format);
    }

    return content;
  }

  /// Loads the i18n file from Assets folder.
  ///
  /// The i18n file must be in [basePath], named [baseName] +
  /// `.<locale>.<extension>` and must be specified in the `pubspec.yaml` file.
  @override
  Future<String> load({required String? locale}) async => _tryLoad(locale);

  /// Loads the i18n file from Assets folder.
  ///
  /// If the [uri] is not null, then we try to load the i18n file
  /// from the given [uri]. In this case, the [basePath] and the
  /// [baseName] are ignored. And there is no fallback.
  @override
  Future<String> loadFrom(Uri uri) async => _tryLoadUri(uri);
}
